import os

from ruamel.yaml import YAML


class GitContext(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__dict__ = self


def fetch_token(host, options, context):
    if 'token' in options:
        return options['token']

    gh_hosts = os.path.expanduser('~/.config/gh/hosts.yml')
    if os.path.exists(gh_hosts):
        yaml = YAML()
        with open(gh_hosts, 'r') as f:
            gh_hosts = yaml.load(f)
        if host in gh_hosts:
            return gh_hosts[host]['oauth_token']
    
    return context.environ.get('GITHUB_TOKEN')



import os

from typing import Dict, List


class ScaffoldContext(dict):
    notes: List[str]
    todos: List[str]
    environ: Dict[str, str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['steps'] = {}
        self.notes = []
        self.todos = []
        self.environ = {}

    def resolve_package_path(self, path):
        package_dir = self['__package']['path']
        return os.path.realpath(os.path.join(package_dir, path))

    def resolve_target_path(self, path):
        target_dir = self['__target']
        return os.path.realpath(os.path.join(target_dir, path))
    
    def set_step(self, step_id, value):
        self['steps'][step_id] = value



import sys
import click
import logging

from click.decorators import help_option

from x_scaffold import engine
from x_scaffold.context import ScaffoldContext
from x_scaffold.runtime import ScaffoldConsoleRuntime


class AddColorFormatter(logging.Formatter):
    def format(self, record):
        msg = super(AddColorFormatter, self).format(record)
        # Green/Cyan/Yellow/Red/Redder based on log level:
        color = (
            "\033[1;"
            + ("32m", "36m", "33m", "31m", "41m")[
                min(4, int(4 * record.levelno / logging.FATAL))
            ]
        )
        return color + record.levelname + "\033[1;0m: " + msg


@click.command('apply', help='Apply a scaffold')
@click.argument('package')
@click.option('-n', '--name', default='xscaffold', help='The name of the scaffold to apply')
@click.option('-t', '--target', default='.', help='Target directory to apply scaffold')
@click.option('-p', '--parameter', multiple=True, help='Parameters to set in the context')
def apply_cli(package, name, target, parameter):
    context = ScaffoldContext({
        '__target': target
    })
    p: str
    for p in parameter:
        eq_idx = p.index('=')
        parameter_name = p[0:eq_idx]
        parameter_value = p[eq_idx+1:]
        context[parameter_name] = parameter_value
    engine.run(context, {
        'package': package,
        'name': name
    }, ScaffoldConsoleRuntime())


@click.group(context_settings=dict(help_option_names=['-h', '--help']))
@click.option('--log-level', default='info', help='The log level to output')
def cli(log_level):
    stdout_hdlr = logging.StreamHandler(stream=sys.stdout)
    stdout_hdlr.setFormatter(AddColorFormatter())

    logging.root.handlers.clear()

    loglevel_str = log_level.upper()
    loglevel = getattr(logging, loglevel_str)

    stdout_hdlr.setLevel(loglevel)
    logging.root.setLevel(loglevel)
    logging.root.addHandler(stdout_hdlr)
    pass

cli.add_command(apply_cli)

if __name__ == '__main__':
    cli()



import json
import os
import re
from typing import List
from ruamel.yaml import YAML

from x_scaffold.context import ScaffoldContext
from x_scaffold.runtime import ScaffoldRuntime
from ..steps import ScaffoldStep
from ..plugin import ScaffoldPluginContext
from ..rendering import render_options, render_text


def init(context: ScaffoldPluginContext):
    context.add_step("config_write", ConfigStep())
    context.add_step("config_read", ConfigReadStep())


def get_formatter(path, options):
    default_formatter = TextConfigFormatter()

    formatters: List[ConfigFormatter] = [
        YamlConfigFormatter(),
        JsonConfigFormatter(),
        default_formatter
    ]

    for formatter in formatters:
        if formatter.is_formatter(path, options):
            return formatter

    return default_formatter


class ConfigStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)

        # target_base = os.path.realpath(context.get('__target', '.'))
        # path: str = os.path.join(target_base, options.get('path'))
        path: str = context.resolve_target_path(options.get('path'))
        configuration = options.get('configuration')

        formatter = get_formatter(path, options)
        formatter.configure(path, configuration)


class ConfigReadStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)
        # target_base = os.path.realpath(context.get('__target', '.'))
        # path: str = os.path.join(target_base, options.get('path'))
        path: str = context.resolve_target_path(options.get('path'))

        formatter = get_formatter(path, options)
        return formatter.read(path)


class ConfigFormatter:
    def read(self, path):
        pass

    def write(self, path, config):
        pass

    def configure(self, path, config):
        pass

    def is_formatter(self, path, options):
        pass


def merge(source, destination: dict):
    dest = destination
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = dest.setdefault(key, {})
            merge(value, node)
        else:
            dest[key] = value

    return dest


class YamlConfigFormatter(ConfigFormatter):
    def __init__(self) -> None:
        self.yaml = YAML()
        super().__init__()

    def read(self, path):
        yaml = self.yaml
        with open(path, 'r') as fhd:
            yaml_dict = yaml.load(fhd)
        return yaml_dict
    
    def write(self, path, config):
        yaml = self.yaml
        with open(path, 'w') as fhd:
            yaml.dump(config, fhd)

    def configure(self, path, config):
        yaml_dict = self.read(path)

        new_config = merge(config, yaml_dict)

        self.write(path, new_config)
    
    def is_formatter(self, path, options):
        format = options.get('format')
        if format == 'yaml' or re.match(r'.*\.y(a)?ml', path):
            return True


class JsonConfigFormatter(ConfigFormatter):
    def __init__(self) -> None:
        super().__init__()

    def read(self, path):
        with open(path, 'r') as fhd:
            yaml_dict = json.load(fhd)
        return yaml_dict
    
    def write(self, path, config):
        with open(path, 'w') as fhd:
            json.dump(config, fhd)

    def configure(self, path, config):
        yaml_dict = self.read(path)

        new_config = merge(config, yaml_dict)

        self.write(path, new_config)
    
    def is_formatter(self, path, options):
        format = options.get('format')
        if format == 'json' or re.match(r'.*\.json', path):
            return True



from x_scaffold.steps import ScaffoldStep
from github import Github, Repository, PullRequest

import os
from ruamel.yaml import YAML

from ..plugin import ScaffoldPluginContext
from ..context import ScaffoldContext
from ..runtime import ScaffoldRuntime
from ..rendering import render_options, render_text


def init(context: ScaffoldPluginContext):
    context.add_step('github_repository', GithubRepoStep())
    context.add_step('github_organization', GithubOrgStep())
    context.add_step('github', GithubStep())


def invokeGH(obj, attribute: str, *args, **kwargs):
    return getattr(obj, attribute)(*args, **kwargs)


def fetch_token(options, context):
    if 'token' in options:
        return options['token']

    host = options.get('host', 'github.com')
    gh_hosts = os.path.expanduser('~/.config/gh/hosts.yml')
    if os.path.exists(gh_hosts):
        yaml = YAML()
        with open(gh_hosts, 'r') as f:
            gh_hosts = yaml.load(f)
        if host in gh_hosts:
            return gh_hosts[host]['oauth_token']
    
    return context.environ.get('GITHUB_TOKEN')


class GithubOrgStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)
        name = options['name']
        token = fetch_token(options, context)
        g = Github(token)
        org = g.get_organization(name)
        context.set_step(step['__id'], org)
        gh_steps = options.get('steps', [])
        gh_step: dict
        for gh_step in gh_steps:
            for k, v in gh_step.items():
                if 'if' in gh_step:
                    if render_text(gh_step['if']) == False:
                        break
                if k not in ['id', 'if']:
                    result = invokeGH(org, k, **v)
                    if 'id' in gh_step:
                        context.set_step(gh_step['id'], result)
        return org


class GithubStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)
        token = fetch_token(options, context)
        g = Github(token)
        context.set_step(step['__id'], g)
        gh_steps = options.get('steps', [])
        for gh_step in gh_steps:
            for k, v in gh_step.items():
                if k in ['id', 'if']:
                    continue
                result = invokeGH(g, k, **v)
                if 'id' in gh_step:
                    context.set_step(gh_step['id'], result)
                break
        return g


# https://pygithub.readthedocs.io/en/latest/github_objects/Repository.html
class GithubRepoStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)
        name = options['name']
        token = fetch_token(options, context)
        g = Github(token)
        repo = g.get_repo(name)
        context.set_step(step['__id'], repo)
        gh_steps = options.get('steps', [])

        for gh_step in gh_steps:
            for k, v in gh_step.items():
                if k in ['id', 'if']:
                    continue
                if 'set_topics' in gh_step:
                    result = self.set_topics(repo, gh_step['set_topics'], context)
                elif 'add_topics' in gh_step:
                    result = self.add_topics(repo, gh_step['add_topics'], context)
                elif 'branch' in gh_step:
                    result = self.run_branch(repo, gh_step['branch'], context)
                else:
                    result = invokeGH(repo, k, **v)
                if 'id' in gh_step:
                    context.set_step(gh_step['id'], result)
                break
        return repo

    def run_branch(self, repo: Repository, step, context):
        name = step['name']
        branch = repo.get_branch(name)
        gh_steps = step.get('steps', [])
        for gh_step in gh_steps:
            for k, v in gh_step.items():
                invokeGH(branch, k, **v)
        return branch
    
    def set_topics(self, repo: Repository, step, context):
        new_topics = step
        repo.replace_topics(new_topics)

    def add_topics(self, repo: Repository, step, context):
        topics = repo.get_topics()
        new_topics = step
        for topic in new_topics:
            if topic not in topics:
                topics.append(topic)
        repo.replace_topics(topics)
    
    def create_pr(self, repo: Repository, step, context):
        opts = render_options(step, context)
        pr: PullRequest.PullRequest = repo.create_pull(**opts)
        context['pr'] = {
            'number': pr.number,
            'url': pr.html_url
        }
    
    def create(self, name: str, github: Github, step, context):
        opts = render_options(step, context)
        parts = name.split('/')
        organization = parts[0]
        name = parts[1]

        github.get_organization(organization).create_repo(name=name, **opts)



import importlib
import importlib
from importlib import util
import subprocess
import sys
import os

from x_scaffold.context import ScaffoldContext
from x_scaffold.runtime import ScaffoldRuntime
from ..steps import ScaffoldStep
from ..plugin import ScaffoldPluginContext
from ..rendering import render_options, render_text


def init(context: ScaffoldPluginContext):
    context.add_step("set_context", SetStep())
    context.add_step("add_note", AddNoteStep())
    context.add_step("add_todo", AddTodoStep())
    context.add_step("module", ModuleStep())


class SetStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        context_names = step
        for context_name in context_names:
            context[context_name] = render_text(
                context_names[context_name], context)


class AddNoteStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: str, runtime: ScaffoldRuntime):
        message = render_text(step, context)
        context.notes.append(message)


class AddTodoStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: str, runtime: ScaffoldRuntime):
        message = render_text(step, context)
        context.todos.append(message)


class ModuleStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)

        install_dependencies(options.get('dependencies', []))

        file = context.resolve_package_path(options['path'])
        spec = util.spec_from_file_location("module.name", file)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        main_fn = getattr(foo, 'main')

        return main_fn(context, options, runtime)


def install_dependencies(dependencies):
    for dep in dependencies:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', dep])



#!/usr/bin/env python
# coding: utf-8

import collections
import logging
import os
import re
import sys
import tempfile

from ruamel.yaml import YAML

from .context import ScaffoldContext
from .plugin import ScaffoldPluginContext
from .plugins import load_plugins
from .rendering import render_options, render_text
from .runtime import ScaffoldRuntime


_log = logging.getLogger(__name__)


# def complete(text, state):
#     if str(text).startswith('~/'):
#         home = os.path.expanduser('~/')
#         p = os.path.join(home, text[2:])
#     else:
#         p = text
#         home = None

#     items = pathlib.Path(os.getcwd()).glob(p + '*')
#     if items is not None and home is not None:
#         items = ['~/' + x[len(home):] for x in items]
#     return (items + [None])[state]


# def set_readline():
#     try:
#         import readline
#         readline.set_completer_delims(' \t\n;')
#         readline.parse_and_bind("tab: complete")
#         readline.set_completer(complete)
#     except:
#         pass


# class AttributeDict(dict):
#     __getattr__ = dict.__getitem__
#     __setattr__ = dict.__setitem__


# color = AttributeDict({
#     'PURPLE': '\033[35m',
#     'CYAN':  '\033[36m',
#     'BLUE':  '\033[34m',
#     'GREEN':  '\033[32m',
#     'YELLOW':  '\033[33m',
#     'RED':  '\033[31m',
#     'BOLD':  '\033[1m',
#     'UNDERLINE':  '\033[4m',
#     'ITALIC':  '\033[3m',
#     'END':  '\033[0m',
# })


# def dict_to_str(d, fmt='%s=%s\n'):
#     s = ''
#     for x in d:
#         s += fmt % (x, d[x])
#     return s


def str2bool(v):
    if v is None:
        return False
    return v.lower() in ("yes", "true", "t", "1", "y")


known_types = {
    'int': int,
    'bool': str2bool,
    'str': str,
    'float': float
}


# def term_color(text, *text_colors):
#     return ''.join(text_colors) + text + color.END




# def render_file(path, context):
#     """Used to render a Jinja template."""

#     template_dir, template_name = os.path.split(path)
#     return render(template_name, context, template_dir)


# def is_enabled(options):
#     if 'enabled' in options:
#         return options['enabled']
#     if 'disabled' in options:
#         return not options['disabled']
#     if 'enabledif' in options:
#         enabledif = options['enabledif']
#         value = enabledif['value']
#         if 'equals' in enabledif:
#             return value == enabledif['equals']
#         elif 'notequals' in enabledif:
#             return value != enabledif['notequals']
#     return True


# def read_input(s):
#     return input(s)


def convert(v, type):
    if type in known_types:
        return known_types[type](v)
    return str(v)


def read_parameter(prompt, context, runtime: ScaffoldRuntime):
    default = prompt.get('default', '')
    # if isinstance(default, str):
    #     default = default.format(env=os.environ)

    #name = prompt.get('name', 'parameter')
    required = prompt.get('required', False)

    if 'if' in prompt:
        enabled = prompt['if']
        if not enabled:
            return default
    
    while True:
        d = runtime.ask(prompt)

        if d == '' or d is None:
            if not required:
                return default
            else:
                runtime.write('{RED}[required]{END} ')
        else:
            if 'validate' in prompt:
                matches = re.match(prompt['validate'], d)
                if matches is None:
                    runtime.write('{RED}[invalid, %s]{END} ' % prompt['validate'])
                    continue
            return convert(d, prompt.get('type', 'str'))


def config_cli(args):
    options = {}
    scaffold_file = os.path.expanduser('~/.xscaffold')

    yaml = YAML()
    if os.path.exists(scaffold_file):
        with open(scaffold_file, 'r') as fhd:
            options = yaml.load(fhd)

    if args.action == 'save':
        options['url'] = args.url

        with open(scaffold_file, 'w') as fhd:
            yaml.dump(options, fhd, default_flow_style=False)
    elif args.action == 'view':
        sys.stdout.write('url: %s' % options.get('url', 'not defined'))


def rm_rf(d):
    for path in (os.path.join(d, f) for f in os.listdir(d)):
        if os.path.isdir(path):
            rm_rf(path)
        else:
            os.unlink(path)
    os.rmdir(d)


def locate_scaffold_file(path, name):
    base_paths = [
        path,
        os.path.join(path, '.xscaffold'),
        os.path.join(path, 'xscaffold')
    ]

    extensions = [
        '.yaml',
        '.yml',
        '.json',
        ''
    ]

    names = [
        f'.{name}',
        f'{name}',
        f'{name}/xscaffold',
        f'{name}/.xscaffold'
    ]
    for base_path in base_paths:
        for ext in extensions:
            for n in names:
                full_path = os.path.join(base_path, n + ext)
                if os.path.exists(full_path):
                    return full_path
    return None


def process_parameters(parameters, context: ScaffoldContext, runtime: ScaffoldRuntime):
    for parameter in parameters:
        parameter_options = render_options(parameter, context)
        parameter_name = parameter_options['name']
        if parameter_name in context:
            context[parameter_name] = convert(context[parameter_name], parameter.get('type', 'str'))
        else:
            context[parameter_name] = read_parameter(parameter_options, context, runtime)


def run(context: ScaffoldContext, options, runtime: ScaffoldRuntime):
    execute_scaffold(context, options, runtime)

    runtime.print_todos(context)
    runtime.print_notes(context)

    return context


def execute_scaffold(context: ScaffoldContext, options, runtime: ScaffoldRuntime):
    tempdir = options.get('temp', tempfile.gettempdir())
    package = options['package']

    yaml = YAML()

    name = options.get('name', 'xscaffold')

    if '__package' in context:
        package_path = context.resolve_package_path(package)
    else:
        package_path = package
    if os.path.exists(package_path):
        _log.debug('using local package \'%s\'', package_path)
        pkg_dir = package_path
    else:
        pkg_dir = fetch_git(runtime, tempdir, package)
    
    scaffold_file = locate_scaffold_file(pkg_dir, name)
    pkg_dir = os.path.dirname(scaffold_file)
    sys.path.append(pkg_dir)
    _log.debug('scaffold file: %s', scaffold_file)

    if scaffold_file is not None:
        with open(scaffold_file, 'r') as fhd:
            config = yaml.load(fhd)
    else:
        config = {
            'steps': options.get('steps', [{ 'fetch': {} }])
        }

    plugin_context = ScaffoldPluginContext(
        config.get('plugins', {})
    )
    plugins: list = load_plugins()
    for plugin in plugins:
        plugin.init(plugin_context)

    context.update(config.get('context', {}))

    process_parameters(config.get('parameters', []), context, runtime)

    context['__package'] = {
        'path': pkg_dir,
        'options': options
    }

    steps_context = context['steps'] = {}
    steps: list = config.get('steps', [])

    execute_steps(context, runtime, plugin_context, steps_context, steps)

    return context

def execute_steps(context: ScaffoldContext, runtime, plugin_context, steps_context, steps):
    step: dict
    for step in steps:
        if 'if' in step:
            enabled = render_text(step['if'], context)
            if enabled == False:
                continue
        if 'group' in step:
            group_steps = step['group']
            execute_steps(context, runtime, plugin_context, steps_context, group_steps)
        elif 'foreach' in step:
            foreach_steps = step['foreach']
            items = render_text(foreach_steps.get('items', []), context)
            step_id = step.get('id', 'foreach')
            for item in items:
                context.set_step(step_id, item)
                execute_steps(context, runtime, plugin_context, steps_context, foreach_steps.get('steps', []))
        else:
            plugin_step_name = None
            for step_name in step:
                if step_name in plugin_context.steps:
                    plugin_step_name = step_name
                    break
            if plugin_step_name:
                plugin_step = plugin_context.steps[plugin_step_name]
                step_options = step[plugin_step_name]

                step_id = step.get('id', plugin_step_name)

                if isinstance(step_options, collections.Mapping):
                    step_options['__id'] = step_id
                
                _log.debug(f'[{step_id}] running')
                result = plugin_step.run(context, step_options, runtime)
                context.set_step(step_id, result)

def fetch_git(runtime, tempdir, package):
    package_parts = package.split('@')
    if len(package_parts) == 1:
        package_name = package_parts[0]
        package_version = 'main'
    else:
        package_name = package_parts[0]
        package_version = package_parts[1]
    package_name_parts = package_name.split('/')
    if len(package_name_parts) <= 2:
        package_name_parts = ['github.com'] + package_name_parts
        package_name = '/'.join(package_name_parts)
    pkg_dir = os.path.join(tempdir, f'{package_name}@{package_version}')

    rc = 9999
    if os.path.exists(pkg_dir):
        runtime.log('{YELLOW}[git] updating %s package...{END}\n' % package)
        rc = os.system(
                """(cd {pkg_dir} && git pull >/dev/null 2>&1)""".format(pkg_dir=pkg_dir))
        if rc != 0:
            runtime.log('{RED}[error]{YELLOW} package %s is having issues, repairing...{END}\n' % package)
            rm_rf(pkg_dir)

    if rc != 0:
        runtime.log('[git] pulling %s package...' % package + '\n')
        rc = os.system(f"""
        git clone https://{package_name} {pkg_dir} >/dev/null 2>&1
        """)
    if rc != 0:
        raise Exception(
                'Failed to pull scaffold package %s' % package)

    rc = os.system(f"""(cd {pkg_dir} && git checkout -f {package_version} >/dev/null 2>&1)""")
    if rc != 0:
        raise Exception('Failed to load version %s' % package_version)
    return pkg_dir
